# Tag Report Utility history, journal, and tag list UI functions

# Copyright 2024 Automation Professionals, LLC <sales@automation-pros.com>
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#   1. Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#   2. Redistributions in binary form must reproduce the above copyright notice,
#      this list of conditions and the following disclaimer in the documentation
#      and/or other materials provided with the distribution.
#   3. The name of the author may not be used to endorse or promote products
#      derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
# SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
# OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
# OF SUCH DAMAGE.

from java.io import ByteArrayOutputStream, PrintWriter
from java.lang import System, String, Long, Boolean, Double, Throwable
from java.nio.charset import StandardCharsets
from java.util import Date
from java.util.concurrent import ConcurrentSkipListMap, LinkedBlockingQueue, TimeUnit, ConcurrentSkipListSet
from java.util.concurrent.atomic import AtomicInteger
from java.util.zip import ZipOutputStream, ZipEntry

from com.inductiveautomation.ignition.common import QualifiedPath, WellKnownPathTypes
from com.inductiveautomation.ignition.common.util import DatasetBuilder, Comparators
from com.inductiveautomation.ignition.common.tags.paths.parser import TagPathParser
from com.inductiveautomation.ignition.common.alarming.config import CommonAlarmProperties, AlarmModeProperties

try:
	from org.apache.commons.text import StringEscapeUtils
except:
	from org.apache.commons.lang3 import StringEscapeUtils

logger = system.util.getLogger(system.util.getProjectName()+'.'+system.reflect.getModulePath())

# Get same provider cache as used/initialized in the providers script.
tagProviderCache = system.util.globalVarMap('tag_provider_folders_cache')
ciAlnumCmp = Comparators.alphaNumeric(False)

# Handle tags selection message.
def tagsSelected(self, payload):
	self.view.custom.idle = False
	self.view.custom.query = payload['query']
	self.view.custom.columns = payload['columns']
	self.view.custom.provider = payload['provider']
	tagList = [x for x in self.view.custom.tags] if self.view.custom.tags else []
	# All inner props are stringified already, coming from Perspective
	pathSet = set(x['fullPath'] for x in tagList)
	for x in payload['results']:
		fpStr = x['fullPath'].toStringFull()
		if not fpStr in pathSet:
			pathSet.add(fpStr)
			tagList.append(x)
	self.view.custom.tags = tagList
	self.view.custom.tallDS = 0
	self.view.custom.wideDS = 0
	self.view.custom.idle = True

# Execute removal of selected tags
def removeTags(self, event):
	removePaths = set(tag.fullPath for tag in self.view.custom.selectedTags)
	prunedTags = [tag for tag in self.view.custom.tags if not tag.fullPath in removePaths]
	self.view.custom.tags = prunedTags
	self.view.custom.tallDS = 0
	self.view.custom.wideDS = 0

# Download list of tag paths, one per line in a text file.
def downloadList(self, event):
	outputName = "%s_TagList_%s.txt" % (system.util.getProjectName(), system.date.format(system.date.now(), 'yyyyMMdd_HHmmss'))
	pathStrings = [tag.fullPath+u"\r\n" for tag in self.view.custom.selectedTags]
	system.perspective.download(outputName, u"".join(pathStrings), 'text/plain; charset=utf-8')

# Tag properties to read after upload in order populate the tag
# export UI properly.
uploadReadProps = ['alarms', 'dataType', 'engUnit', 'formatString', 'historyProvider', 'tagType', 'valueSource']

# Upload and repopulate a list of tag paths, one per line in a text file.
# The file upload field's onFileReceived delegates here.
def uploadList(self, event):
	fullPaths = []
	providers = []
	paths = []
	names = []
	readPaths = []
	for line in event.file.getString().splitlines():
		# Ignore empty lines
		if line and line.strip():
			tp = TagPathParser.parse(line)
			if tp.property:
				continue
			fullPath = tp.toStringFull()
			fullPaths.append(fullPath)
			providers.append(tp.source)
			paths.append(tp.toStringPartial())
			names.append(tp.itemName)
			readPaths += [tp.toStringFull() + '.' + p for p in uploadReadProps]
		else:
			logger.warnf("Discarded blank? '%s'", line)
	# Ready to read all properties
	qvList = system.tag.readBlocking(readPaths)
	logger.debugf("Tag Prop reads=%s", repr(qvList))
	propList = [qv.value for qv in qvList]
	propGroupIter = [propList.__iter__()] * len(uploadReadProps)
	groupedTagProps = map(None, *propGroupIter)
	# Construct final list of tags as dictionaries of properties
	tags = []
	for fp, prov, p, n, props in zip(fullPaths, providers, paths, names, groupedTagProps):
		tag = dict(fullPath = fp, provider = prov, path = p, name = n)
		tag.update([(k, v) for (k, v) in zip(uploadReadProps, props) if v is not None])
		if 'alarms' in tag:
			tag['alarms'] = [dict([(pv.property.getName(), pv.value) for pv in alarmdef if pv.value is not None]) for alarmdef in tag['alarms'].definitions]
		tags.append(tag)
		logger.debugf("uploadList Tag=%s props=%s", repr(tag), repr(props))
	# Replace any prior list
	self.view.custom.tags = tags
	self.view.custom.tallDS = 0
	self.view.custom.wideDS = 0

# Tall history is four columns:
# "#NAMES"
# "path","value","quality","timestamp"
# "#TYPES"
# "str","L","I","date"
# "#ROWS","0"

# Wide history is n+1 columns, where the first column is t_stamp of
# type date. Remaining columns have names matching the tag path and
# tag & aggregation-specific datatype.

# Requested tags are grouped by provider so that the returned paths
# can be unambiguously re-associated with their providers.

# History is post-processed into the pageVarMap().histByProvider, which is
# a dictionary keyed by tag provider, which itself contains dictionaries
# keyed by partial path.  The inner dictionary values are the dataset of
# just that path's history.  This grouped history has consecutive null values
# consolidated (all but the first are dropped).

alarmEventTablesSQL = """Select c1.table_schema, c1.table_name
From information_schema.columns c1
Inner Join information_schema.columns c2 ON c1.table_schema=c2.table_schema AND c1.table_name=c2.table_name AND c2.column_name = 'eventid'
Inner Join information_schema.columns c3 ON c1.table_schema=c3.table_schema AND c1.table_name=c3.table_name AND c3.column_name = 'source'
Inner Join information_schema.columns c4 ON c1.table_schema=c4.table_schema AND c1.table_name=c4.table_name AND c4.column_name = 'displaypath'
Inner Join information_schema.columns c5 ON c1.table_schema=c5.table_schema AND c1.table_name=c5.table_name AND c5.column_name = 'priority'
Inner Join information_schema.columns c6 ON c1.table_schema=c6.table_schema AND c1.table_name=c6.table_name AND c6.column_name = 'eventtype'
Inner Join information_schema.columns c7 ON c1.table_schema=c7.table_schema AND c1.table_name=c7.table_name AND c7.column_name = 'eventflags'
Inner Join information_schema.columns c8 ON c1.table_schema=c8.table_schema AND c1.table_name=c8.table_name AND c8.column_name = 'eventtime'
where c1.column_name = 'id' 
"""

alarmOutColumns = ['t_stamp', 'UUID', 'Type', 'Path', 'Alarm', 'DisplayPath', 'State']
alarmOutTypes = [Date, String, String, String, String, String, String]

alarmSQL = """
SELECT e."eventtime" AS "t_stamp",
	e."eventid" AS "uuid",
	CASE e.eventtype
		WHEN 0 THEN 'Active'
		WHEN 1 THEN 'Clear'
		WHEN 2 THEN 'Ack'
		WHEN 3 THEN 'Enable'
		ELSE 'Disable'
	END AS "evtype",
	e."source",
	e."displaypath",
	CASE e."eventflags" & 24
		WHEN 0 THEN 'Active Unacked'
		WHEN 8 THEN 'Active Acked'
		WHEN 16 THEN 'Clear Unacked'
		ELSE 'Clear Acked'
	END AS "state"
FROM "%s"."%s" e
WHERE e."eventtime" >= ? AND e."eventtime" < ?
  AND e."source" IN (%s)
ORDER BY e."eventtime"
"""

def coalesce(*args):
	for x in args:
		if x is not None:
			return x
	return None
#

# Tag config is a dictionary-like object describing a tag, as returned
# by system.tag.query() or similar methods.  The provider, storageProvider,
# and path properties interact with the system's configured tag provider
# and history provider names, with special "driver" handling for remote
# tag providers.

# Given a tag config dictionary, start by looking up the provider in the
# provider cache. If missing, just return the fullpath value, and let the
# history api deal with it.

# Then check if the history provider name exists in the list of
# historians. If present, construct and return a qualified path from the
# history provider, the cached driver, and the tag path.

# Otherwise, return the fullpath value and let the
# history API interpret as appropriate for remote or disabled
# historians.
def histPathFromTagConfig(cfg):
	cachedProvider = tagProviderCache.get(cfg['provider'])
	if cachedProvider:
		if cfg['historyProvider'] in tagProviderCache.get('!!Hist', []):
			return QualifiedPath.Builder().set(WellKnownPathTypes.HistoryProvider, cfg['historyProvider']).setDriver(cachedProvider['driver']).setTag(cfg['path']).build()
	return cfg['fullPath']

# If multiple tag providers are to be queried, they will run in parallel in
# a temporary thread pool.
#
# Output is to pvm.histByProv[provider].tall or .wide, and .byPath[path]
def parallelHistory(pvm, kw, provider, selection):
	kw['paths'] = [histPathFromTagConfig(x) for x in selection]
	ds = system.tag.queryTagHistory(**kw)
	prunedKW = dict(kw)
	prunedKW.pop('paths')
	logger.infof("History query for %s arguments %s, %d tags => %s", provider, prunedKW, len(kw.get('paths', [])), ds)
	pResult = pvm.histByProv[provider]
	pResult.all = ds
	if kw['returnFormat'] == 'Tall':
		builderMap = {}
		for row in system.dataset.toPyDataSet(ds):
			p, v, ts = [row[c] for c in ['path', 'value', 'timestamp']]
			dsb = builderMap.get(p)
			if dsb is None:
				dsb = builderMap.setdefault(p, DatasetBuilder().colNames(['t_stamp', p]).colTypes([Date, Double]))
			dsb.addRow([ts, v])
		for p, dsb in builderMap.items():
			pResult.byPath[p] = dsb.build()
	else:
		for c in range(1, ds.columnCount):
			p = ds.getColumnName(c)
			pResult.byPath[p] = system.dataset.filterColumns(ds, [0, c])

# Helper for alarm result datasets
def makeAlarmOutBuilder():
	return DatasetBuilder().colNames(alarmOutColumns).colTypes(alarmOutTypes)

# The combination of all tag providers and all possible alarm journal
# tables will be queried for corresponding alarms.  If multiple (likely),
# they will run in parallel in a temporary thread pool.  (The same thread
# pool that starts with the historian queries.)
#
# Output is to pvm.alarmsByProv[provider].fragments
def parallelJournal(pvm, db, schema, table, provider, preciseStart, preciseEnd, alarmTags):
	pResult = pvm.alarmsByProv[provider]
	prefix = "prov:" + provider + ":/tag:"
	alarmPaths = [prefix + x['path'] + ":/alm:" + y['name'] for x in alarmTags for y in x['alarms']]
	dsb0 = makeAlarmOutBuilder()
	while alarmPaths:
		# Limit queries to 2000 replacement parameters. First two are the timestamps.
		alarmPartialPaths = alarmPaths[:1998]
		del alarmPaths[:1998]
		sql = alarmSQL % (schema, table, ', '.join(['?'] * len(alarmPartialPaths)))
		params = [preciseStart, preciseEnd] + alarmPartialPaths
		pyds = system.db.runPrepQuery(sql, params, db)
		logger.infof("Journal query for %s db=% schema=%s table=%s %d paths => %s",
			provider, db, schema, table, len(alarmPartialPaths), pyds)
		for row in pyds:
			qp = QualifiedPath.parse(row[3])
			path = qp.getPathComponent('tag')
			alarm = qp.getPathComponent('alm')
			newRow = row[:3] + [path, alarm] + row[4:]
			dsb0.addRow(newRow)
	pResult.fragments.append((db, schema, table, dsb0.build()))

# Constant for sorting and deduplication
alarmUniqueness = ['t_stamp', 'Path', 'Alarm', 'Type']

# When all parallel journal queries are complete, the intermediate results
# in pvm.alarmsByProv[provider].fragments must be merged, sorted, and
# deduplicated.  The deduplicated results are simultaneously added to
# the overall provider dataset builder and the appropriate per-path
# builder.
#
# Output is to pvm.alarmsByProv[provider].all and to .byPath[path]
def parallelDedup(pvm, provider, alarmTags):
	pResult = pvm.alarmsByProv[provider]
	allDS = None
	for db, schema, table, fragmentDS in pResult.fragments:
		if fragmentDS is None:
			continue
		allDS = fragmentDS if allDS is None else system.dataset.appendDataset(allDS, fragmentDS)
	del pResult['fragments']
	if allDS is None:
		return
	orderedDS = system.dataset.orderBy(allDS, *alarmUniqueness)
	dsb0 = makeAlarmOutBuilder()
	builderMap = dict([(cfg['path'], makeAlarmOutBuilder()) for cfg in alarmTags])
	prior = [None] * len(alarmUniqueness)
	for row in system.dataset.toPyDataSet(orderedDS):
		uniques = [row[c] for c in alarmUniqueness]
		if prior == uniques:
			continue
		prior = uniques
		bareRow = [x for x in row]
		dsb0.addRow(bareRow)
		path = uniques[1]
		dsb = builderMap.get(path)
		if dsb:
			dsb.addRow(bareRow)
	pResult.all = dsb0.build()
	logger.infof("Journal deduplication for %s => %s",
			provider, pResult.all)
	for p, dsb in builderMap.items():
		pResult.byPath[p] = dsb.build()

# History retrieval using the time span from the view custom properties and
# the tag selection results in view.custom.selectedTags.
# Results are delivered through the pageVarMap() under the keys
# 'histByProv' and 'alarmsByProv'.
# Each of the above uses realtime tag provider names as keys, leading to a
# VarMap instance.

# Under .histByProv[provider], either the .tall or the .wide key receives the
# complete history dataset for that provider.  The .byPath key receives
# a dictionary, keyed by simple tagpath, containing the content for just
# that tag. (Two columns, t_stamp and value, but the value column name is
# the tag path, as if unconditionally wide-format.

# Under .alarmsByProv[provider], the .all key receives the de-duplicated
# content for all alarm records found across all data sources and all
# tables having the structure of alarm_events.  The .byPath key receives
# a dictionary, keyed by simple tagpath, containing the content for just
# that tag.

# Because the alarm queries must be de-duplicated after *all* data is
# retrieved, the code below creates parallel threads for that post-
# processing after all query threads are complete.

# The .byPath dictionaries are used in the tag display UI and the charting
# UI, particularly by the main table's cell views.  This permits the main
# UI table to have static content after the tag list is set up, avoiding
# table UI annoyances.  (Deselection on fresh data, especially.)

def getSelectedHistory(self):
	logger.info("Starting History and Alarm Queries")
	try:
		pvm = system.perspective.pageVarMap(self)
		self.view.custom.idle = False
		self.view.custom.stale = False
		queue = LinkedBlockingQueue()
		# History Queries by Provider
		# Assemble the keyword arguments that apply to all providers.
		kw = {}
		preciseStart = exchange.tagReportUtility.timespan.coerceDate(self.view.custom.preciseStart)
		kw['startDate'] = preciseStart
		preciseEnd = exchange.tagReportUtility.timespan.coerceDate(self.view.custom.preciseEnd)
		kw['endDate'] = system.date.addMillis(preciseEnd, -1)
		if self.view.custom.intervals.modulus > 1:
			kw['aggregationMode'] = self.view.custom.aggregation
			kw['returnSize'] = (preciseEnd.time - kw['startDate'].time) / self.view.custom.intervals.modulus
			kw['returnFormat'] = 'Wide'
			kw['interval'+self.view.custom.intervals.increment] = self.view.custom.intervals.divisor
			tall = False
		else:
			kw['returnSize'] = -1
			kw['returnFormat'] = 'Tall'
			tall = True
		# Filter the tag selection to just the historized tags.
		selectedHistTags = [x for x in self.view.custom.selectedTags if x.get('historyProvider')]
		logger.infof("All history paths: %s", len(selectedHistTags))
		histTagsByProvider = ConcurrentSkipListMap(ciAlnumCmp)
		for cfg in selectedHistTags:
			inner = histTagsByProvider.setdefault(cfg['provider'], [])
			inner.append(cfg)
		logger.infof("Providers to query: %s", repr([x for x in histTagsByProvider]))
		# Initialize result info by provider
		for provider in histTagsByProvider:
			pResult = pvm.setdefault('histByProv', {})[provider] = system.util.VarMap()
			pResult.byPath = ConcurrentSkipListMap(ciAlnumCmp)
		# Prune old provider results that won't be updated
		if pvm.histByProv:
			deletions = [x for x in pvm.histByProv if not x in histTagsByProvider]
			if deletions:
				for x in deletions:
					del pvm.histByProv[x]
		# Queue up the history query work
		for provider, inner in histTagsByProvider.items():
			queue.offer((parallelHistory, [pvm, kw, provider, inner], {}))
		logger.infof("Queued work for providers: %s", repr([x for x in histTagsByProvider]))

		# Alarm Journal Queries via database calls.  NOT via API. (API is
		# brutally slow.)
		# Journal recording criteria are unknown, so all paths will be requested
		# from all journals. Duplicates will be discarded at the end, when all
		# results by provider have been merged into a single dataset and sorted.

		# Filter the tag selection to just the tags with alarms.
		selectedAlarmTags = [x for x in self.view.custom.selectedTags if x.get('alarms', [])]
		alarmTagsByProvider = ConcurrentSkipListMap(ciAlnumCmp)
		for cfg in selectedAlarmTags:
			inner = alarmTagsByProvider.setdefault(cfg['provider'], [])
			inner.append(cfg)
		# Initialize result info by provider
		for provider in alarmTagsByProvider:
			pResult = pvm.setdefault('alarmsByProv', {})[provider] = system.util.VarMap()
			pResult.fragments = []
			pResult.byPath = ConcurrentSkipListMap(ciAlnumCmp)
		# Prune old provider results that won't be updated
		if pvm.alarmsByProv:
			deletions = [x for x in pvm.alarmsByProv if not x in alarmTagsByProvider]
			if deletions:
				for x in deletions:
					del pvm.alarmsByProv[x]
		# Construct lists of alarm events tables by datasource, then flatten
		# so work will be queued to alternating datasources.  Each final entry is a
		# tuple of DB, Schema, and Table
		databases = system.db.getConnections()
		tablesByDatabase = []
		for dbrow in system.dataset.toPyDataSet(databases):
			if dbrow['Status'] == 'Valid':
				db = dbrow['Name']
				tablesDS = system.db.runQuery(alarmEventTablesSQL, db)
				tables = [(db, row[0], row[1]) for row in system.dataset.toPyDataSet(tablesDS)]
				# Skip when no alarm_events tables in the database
				if tables:
					tablesByDatabase.append(tables)
		# Flatten
		if len(tablesByDatabase) > 1:
			# flatten with map()
			flatTables = [x for row in map(None, *tablesByDatabase) for x in row if x]
		elif tablesByDatabase:
			# Just one.  Unwrap it.
			flatTables = tablesByDatabase[0]
		else:
			flatTables = []
		# Queue up the alarm journal database work
		for provider, alarmTags in alarmTagsByProvider.items():
			for db, schema, table in flatTables:
				queue.offer((parallelJournal, [pvm, db, schema, table, provider, preciseStart, preciseEnd, alarmTags], {}))

		# Run all the queries
		workCounter = AtomicInteger()
		def processQueue():
			try:
				# queue and folderCounter are captured by the closure
				preCount = -1
				while True:
					task = queue.poll(25L, TimeUnit.MILLISECONDS)
					if task:
						f, args, kwargs = task
						count = workCounter.incrementAndGet()
						try:
							f(*args, **kwargs)
						except:
							pass
					else:
						count = workCounter.get()
					if count == preCount:
						break
					preCount = count
			except Throwable, t:
				logger.warn("Unexpected java error processing task", t)
			except Exception, e:
				logger.warn("Unexpected jython error processing task", system.reflect.asThrowable(e, None))
		# make and join threads
		threadList = [system.util.invokeAsynchronous(processQueue, description="History/Journal worker #%d" % (x+1)) for x in range(4)]
		for th in threadList:
			th.join()

		# Queue up the work to post-process alarm results by provider
		for provider, alarmTags in alarmTagsByProvider.items():
			queue.offer((parallelDedup, [pvm, provider, alarmTags], {}))
		# make and join threads
		threadList = [system.util.invokeAsynchronous(processQueue, description="History/Journal worker #%d" % (x+1)) for x in range(4)]
		for th in threadList:
			th.join()
	except Throwable, t:
		logger.warn("Unexpected java error in history/alarm queries", t)
	except Exception, e:
		logger.warn("Unexpected jython error in history/alarm queries", system.reflect.asThrowable(e, None))
	finally:
		self.view.custom.idle = True
		pvm.refresh()
#

# Called in two places.
def autoSize(sheet, qty):
	for c in range(qty):
		sheet.autoSizeColumn(c)
		sheet.setColumnWidth(c, sheet.getColumnWidth(c) + 256)

# Dataset to single sheet.  Top left cell is string description.
# Second row is column headers, but added last (to not be auto-sized).
def datasetToSheet(wb, sheet, hStyle, rStyle, dStyle, ds, title):
	titleRow = sheet.createRow(0)
	headerRow = sheet.createRow(1)
	# Populate later.  Set up for auto-sizing.
	doAutoSize = True
	for r in range(ds.rowCount):
		row = sheet.createRow(r + 2)
		for c in range(ds.columnCount):
			value = ds.getValueAt(r, c)
			if not value is None:
				cell = row.createCell(c)
				if isinstance(value, Date):
					cell.setCellStyle(dStyle)
				else:
					cell.setCellStyle(rStyle)
				cell.setCellValue(value)
		if r == 50:
			# When more than fifty rows, do the autosize early, as it
			# can be very time-consuming.
			doAutoSize = False
			autoSize(sheet, ds.columnCount)
	if doAutoSize:
		autoSize(sheet, ds.columnCount)
	# Now populate the title row.  It is expected to be long, and can
	# overflow to adjacent columns.
	titleCell = titleRow.createCell(0)
	titleCell.setCellStyle(hStyle)
	titleCell.setCellValue(title)
	# Now populate the header row.  Likely to be long tag paths, so not
	# auto-sized.
	for c, name in enumerate(ds.columnNames):
		cell = headerRow.createCell(c)
		cell.setCellStyle(hStyle)
		cell.setCellValue(name)

# Perform a dowload of the acquired data, in Excel format
def excelDownload(self, event):
	pvm = system.perspective.pageVarMap(self)
	wb, hStyle, rStyle, dStyle, jStyle = exchange.tagReportUtility.excel.createWorkbook()
	self.view.custom.idle = False
	try:
		span = system.date.format(self.view.custom.preciseStart, 'yyyy-MM-dd HH:mm:ss.SSS') + u' ≤ ts < ' + system.date.format(self.view.custom.preciseEnd, 'yyyy-MM-dd HH:mm:ss.SSS z')
		outputName = "%s_TagHistory_%s.xlsx" % (system.util.getProjectName(), system.date.format(system.date.now(), 'yyyyMMdd_HHmmss'))
		if pvm.alarmsByProv:
			for provider, pResult in pvm.alarmsByProv.items():
				if self.view.custom.sheetsByPath:
					i = 1
					for path, result in pResult.byPath.items():
						if result and result.rowCount:
							sheet = wb.createSheet("A%03d %s" % (i, provider))
							datasetToSheet(wb, sheet, hStyle, rStyle, dStyle, result, "Alarms for [%s]%s %s" % (provider, path, span))
							i+=1
				else:
					if pResult.all and pResult.all.rowCount:
						sheet = wb.createSheet("A %s" % provider)
						datasetToSheet(wb, sheet, hStyle, rStyle, dStyle, pResult.all, "Alarms for [%s] %s" % (provider, span))
		if pvm.histByProv:
			for provider, pResult in pvm.histByProv.items():
				if self.view.custom.sheetsByPath:
					i = 1
					for path, result in pResult.byPath.items():
						if result and result.rowCount:
							sheet = wb.createSheet("H%03d %s" % (i, provider))
							datasetToSheet(wb, sheet, hStyle, rStyle, dStyle, result, "History for [%s]%s %s" % (provider, path, span))
							i+=1
				else:
					if pResult.all and pResult.all.rowCount:
						sheet = wb.createSheet("H %s" % provider)
						datasetToSheet(wb, sheet, hStyle, rStyle, dStyle, pResult.all, "History for [%s] %s" % (provider, span))
		exchange.tagReportUtility.excel.downloadWorkbook(wb, outputName)
	finally:
		self.view.custom.idle = True
#

# Custom CSV output with minimal quoting.
def formatForCSV(val):
	if isinstance(val, Date):
		return system.date.format(val, 'yyyy-MM-dd HH:mm:ss.SSS')
	v = unicode(val)
	escaped = StringEscapeUtils.escapeJava(v)
	if v == escaped:
		return v
	return '"' + escaped + '"'
#

# Write dataset to ZipOutputStream
def datasetToZip(zos, name, ds, title):
	zos.putNextEntry(ZipEntry(name + '.csv'))
	baos = ByteArrayOutputStream()
	pw = PrintWriter(baos, False, StandardCharsets.UTF_8)
	pw.println("# "+title)
	pw.println(",".join([formatForCSV(x) for x in ds.columnNames]))
	for row in system.dataset.toPyDataSet(ds):
		pw.println(",".join([formatForCSV(x) for x in row]))
	pw.flush()
	zos.write(baos.toByteArray())
	zos.closeEntry()
#

# Perform a dowload of the acquired data, in Zipped CSV format
def zipDownload(self, event):
	pvm = system.perspective.pageVarMap(self)
	self.view.custom.idle = False
	try:
		span = system.date.format(self.view.custom.preciseStart, 'yyyy-MM-dd HH:mm:ss.SSS') + u' ≤ ts < ' + system.date.format(self.view.custom.preciseEnd, 'yyyy-MM-dd HH:mm:ss.SSS z')
		outputName = "%s_TagHistory_%s.zip" % (system.util.getProjectName(), system.date.format(system.date.now(), 'yyyyMMdd_HHmmss'))
		baos = ByteArrayOutputStream()
		zos = ZipOutputStream(baos)
		if pvm.alarmsByProv:
			for provider, pResult in pvm.alarmsByProv.items():
				if self.view.custom.sheetsByPath:
					i = 1
					for path, result in pResult.byPath.items():
						if result and result.rowCount:
							innerName = "A_%s_%d" % (system.dataset.mungeColumnName(provider), i)
							datasetToZip(zos, innerName, result, "Alarms for [%s]%s %s" % (provider, path, span))
							i+=1
				else:
					if pResult.all and pResult.all.rowCount:
						innerName = "A_%s" % system.dataset.mungeColumnName(provider)
						datasetToZip(zos, innerName, pResult.all, "Alarms for [%s] %s" % (provider, span))
		if pvm.histByProv:
			for provider, pResult in pvm.histByProv.items():
				if self.view.custom.sheetsByPath:
					i = 1
					for path, result in pResult.byPath.items():
						if result and result.rowCount:
							innerName = "H_%s_%d" % (system.dataset.mungeColumnName(provider), i)
							datasetToZip(zos, innerName, result, "History for [%s]%s %s" % (provider, path, span))
							i+=1
				else:
					if pResult.all and pResult.all.rowCount:
						innerName = "H_%s" % system.dataset.mungeColumnName(provider)
						datasetToZip(zos, innerName, pResult.all, "History for [%s] %s" % (provider, span))
		zos.close()
		system.perspective.download(outputName, baos.toByteArray(), 'application/csv')
	finally:
		self.view.custom.idle = True
#

#
# kate: tab-width 4; indent-width 4; tab-indents on; dynamic-word-wrap off; indent-mode python; line-numbers on;
