# Tag Report Utility provider caching of names, folders, and properties.

# Copyright 2024 Automation Professionals, LLC <sales@automation-pros.com>
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#   1. Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#   2. Redistributions in binary form must reproduce the above copyright notice,
#      this list of conditions and the following disclaimer in the documentation
#      and/or other materials provided with the distribution.
#   3. The name of the author may not be used to endorse or promote products
#      derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
# SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
# OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
# OF SUCH DAMAGE.

from threading import RLock

from java.lang import System
from java.util import Locale
from java.util.concurrent import ConcurrentSkipListMap, LinkedBlockingQueue, TimeUnit
from java.util.concurrent.atomic import AtomicInteger
from com.inductiveautomation.ignition.common.config import BasicDescriptiveProperty
from com.inductiveautomation.ignition.common.i18n import LocalizedString
from com.inductiveautomation.ignition.common.util import Comparators
from com.inductiveautomation.ignition.common.tags.config.types import TagObjectType
try:
	from com.inductiveautomation.ignition.gateway.tags.distributed.client import RemoteTagProvider
	rgwf = RemoteTagProvider.getDeclaredField('gateway')
	rgwf.accessible = True
	rpf = RemoteTagProvider.getDeclaredField('remoteProvider')
	rpf.accessible = True
	from com.inductiveautomation.ignition.gateway import ContextState
except:
	pass

logger = system.util.getLogger(system.util.getProjectName()+'.'+system.reflect.getModulePath())
tkCtx = system.util.toolkitCtx()

# Cache all tag folders in nested dictionaries to drive the common (custom) tag
# browse tree used by the tag selection popup.  The top level keys are the bare
# provider names.

tagProviderCache = system.util.globalVarMap('tag_provider_folders_cache')
tpcLock = tagProviderCache.setdefault('!!Lock', RLock())
udtIcon = {'path': 'material/folder_special', 'color': "#C0C0C0", "style": {}}
ciAlnumCmp = system.util.alNumCompare(False)

def beanMapOf(obj):
	retv = {}
	for attr in dir(obj):
		if attr.startswith("_"):
			continue
		try:
			value = getattr(obj, attr)
		except:
			continue
		if callable(value):
			continue
		if isinstance(value, LocalizedString):
			retv[attr] = value.toString(Locale.US)
			continue
		retv[attr] = value
	return retv

def getProviderProperties(provider):
	tm = tkCtx.tagManager
	future = tm.getTagConfigModelAsync(provider)
	propModel = future.get()
	props = ConcurrentSkipListMap(ciAlnumCmp)
	for prop in propModel.modelProperties:
		props.put(prop.name, (prop, beanMapOf(prop)))
	#
	# Add custom properties here, using instances of BasicDescriptiveProperty
	#
	return props
#

# Query for provider init based on system.tag.query().
# Also returns UDT definition *folders*.  Prune those by _types_ prefix.

folderAndUdtQuery = {
  "options": {"includeUdtMembers": True, "includeUdtDefinitions": False},
  "condition": {
    "path": "*",
    "attributes": {"values": [], "requireAll": True},
    "properties": {
      "op": "Or",
      "conditions": [
        {"prop": "tagType", "comp": "Equal", "value": "Folder"},
        {"prop": "tagType", "comp": "Equal", "value": "UdtInstance"}
      ]
    }
  },
  "returnProperties": ["tagType"]
}

def makeItem(ftp, isUdt):
	base = dict(
		folders=ConcurrentSkipListMap(ciAlnumCmp),
		udts=ConcurrentSkipListMap(ciAlnumCmp),
		label=ftp.itemName,
		data=ftp.toStringPartial()
	)
	if isUdt:
		base['icon'] = udtIcon
	return base
#

def doProviderInit(providerName, driver):
	logger.infof("Starting browse of provider %s, driver=%s", providerName, driver)
	try:
		thisMap = dict(
			folders=ConcurrentSkipListMap(ciAlnumCmp),
			udts=ConcurrentSkipListMap(ciAlnumCmp),
			driver=driver,
			props=getProviderProperties(providerName)
		)
		tagProviderCache[providerName] = thisMap
	
		results = system.tag.query(providerName, folderAndUdtQuery)
		for tag in results.results:
			ftp = tag['fullPath']
			if ftp.getPathComponent(0) == '_types_':
				continue
			ftplast = ftp.pathLength - 1
			itemMap = thisMap
			for i in range(ftp.pathLength):
				component = ftp.getPathComponent(i)
				if i == ftplast:
					if tag['tagType'] == TagObjectType.Folder:
						itemMap['folders'][component] = makeItem(ftp, False)
					else:
						itemMap['udts'][component] = makeItem(ftp, True)
				else:
					innerMap = itemMap['folders'].get(component)
					if innerMap is None:
						innerMap = itemMap['udts'].get(component)
					if innerMap is None:
						break
					itemMap = innerMap
	except Throwable, t:
		logger.info("Java Error initializing provider "+providerName, t)
	except Exception, e:
		logger.info("Jython Error initializing provider "+providerName, system.reflect.asThrowable(e, None))
#

# Called only from the gateway startup event.  Delays calling doInit() for
# at least 15 seconds after the gateway reaches the running state.
def doConditionalInit():
	if tkCtx.state is ContextState.STARTING:
		tkCtx.executionManager.executeOnce(doConditionalInit, 5000L)
		logger.info("Deferring Tag Provider Cache Init in gateway startup.")
		return
	if tkCtx.state is ContextState.RUNNING:
		if '!!Idle' in tagProviderCache:
			# Already run at least once.  Do not defer further.
			doInit()
		else:
			tkCtx.executionManager.executeOnce(doInit, 15000L)
			logger.info("Deferring Tag Provider Cache Init for provider startup.")
		return
	logger.warnf("Skipping Tag Provider Cache Init in state %s", tkCtx.state)
#

#
@system.util.runInGateway
def getDirectory():
	return dir(tkCtx.systemProperties)
#

# Run all provider browse operations.  Called indirectly from a gateway
# startup event, and directly from the Perspective UI providers refresh
# button.
def doInit():
	with tpcLock:
		logger.info("Starting providers browse")
		startTS = System.nanoTime()
		# Disable UI refresh button when already running
		tagProviderCache['!!Idle'] = False
		tagProviderCache.refresh()
		tagProviderCache['!!Hist'] = tkCtx.tagHistoryManager.tagHistoryProviders
		try:
			tagProviderCache['!!System'] = tkCtx.systemPropertiesManager.systemName
		except:
			# Deprecated as dangerous
			tagProviderCache['!!System'] = tkCtx.systemProperties.systemName
		gtm = tkCtx.tagManager
		# Populate the queue with the task for each provider
		priorProviders = set([k for k in tagProviderCache if not k.startswith("!!")])
		queue = LinkedBlockingQueue()
		for tp in gtm.tagProviders:
			logger.infof("Realtime Tag Provider: [%s] type %s", tp.name, tp.getClass().getSimpleName())
			if isinstance(tp, RemoteTagProvider):
				rgw = rgwf.get(tp)
				rp = rpf.get(tp)
				driver = rgw.gatewayId.serverName + ':' + rp
			else:
				driver = tagProviderCache['!!System'] + ':' + tp.name
			queue.offer((tp.name, driver))
			priorProviders.discard(tp.name)
		# Process the queue with multiple threads.
		def processQueue():
			# Run until queue is empty
			preCount = -1
			task = queue.poll()
			while task:
				providerName, driver = task
				try:
					doProviderInit(providerName, driver)
				except:
					pass
				task = queue.poll()
		# make threads
		threadQty = min(4, len(queue))
		threadList = [system.util.invokeAsynchronous(processQueue, description="Provider browser #%d" % (x+1)) for x in range(threadQty)]
		# Do discards
		for providerName in priorProviders:
			tagProviderCache.pop(providerName)
		# join threads
		for th in threadList:
			th.join()

		finishTS = System.nanoTime()
		logger.infof("Completed providers browse in %.1fms", 0.000001 * (finishTS - startTS))
		tagProviderCache['!!Idle'] = True
		tagProviderCache.refresh()
#
# Designer Script Console accessors

@system.util.runInGateway
def getSystem():
	return tagProviderCache.get('!!System')

@system.util.runInGateway
def getHistorians():
	return tagProviderCache.get('!!Hist')

@system.util.runInGateway
def getProviders():
	return [prov for prov in tagProviderCache if not prov.startswith('!!')]

@system.util.runInGateway
def getProjectProvider():
	proj = system.util.getProjectName()
	pgprops = tkCtx.projectManager.getProjectProps(proj)
	return pgprops.defaultSQLTagsProviderName

#
# kate: tab-width 4; indent-width 4; tab-indents on; dynamic-word-wrap off; indent-mode python; line-numbers on;
