# Tag Report Utility tag selection UI scripts.

# Copyright 2024 Automation Professionals, LLC <sales@automation-pros.com>
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#   1. Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#   2. Redistributions in binary form must reproduce the above copyright notice,
#      this list of conditions and the following disclaimer in the documentation
#      and/or other materials provided with the distribution.
#   3. The name of the author may not be used to endorse or promote products
#      derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
# SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
# OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
# OF SUCH DAMAGE.

# Utility functions needed by the TagSelect popup view
from java.io import FileOutputStream, ByteArrayOutputStream
from java.lang import System, Thread, Throwable
from java.util import Date, Map, ArrayList, UUID
from com.inductiveautomation.ignition.common.model.values import QualifiedValue
from com.inductiveautomation.ignition.common.project import GlobalProps
from com.inductiveautomation.ignition.common.util import ResourceUtil
from com.inductiveautomation.ignition.common.util import Comparators
from org.apache.poi.ss.usermodel import Cell, Row, Sheet, Workbook
from org.apache.poi.xssf.usermodel import XSSFWorkbook, XSSFFont, XSSFDataFormat
import re

try:
	pHook = system.util.toolkitCtx().getModule('com.inductiveautomation.perspective')
	objWrapper = pHook.class.classLoader.loadClass("com.inductiveautomation.perspective.gateway.script.PropertyTreeScriptWrapper$ObjectWrapper")
	mappings = (dict, Map, objWrapper)
except:
	mappings = (dict, Map)

logger = system.util.getLogger(system.util.getProjectName()+'.'+system.reflect.getModulePath())

properAliasRE = re.compile(r'([a-z])([A-Z])')

def properAlias(key):
	key = str(key)
	return key[:1].upper() + properAliasRE.sub(r'\1 \2', key[1:])

uiConstants = {
  # Supply the base list of possible non-exact quality names, and the
  # inner lists of exact quality names that go with them.  In the order
  # to present to the user interface in the selection tree.
  "qualities":
    [
	  {"name": "Good", "exacts": ["Good_Backfill", "Good_Initial", "Good_Overload", "Good_Provisional", "Good_Unspecified", "Good_WritePending"]},
	  {"name": "Bad", "exacts": ["Bad_AccessDenied", "Bad_AggregateNotFound", "Bad_DatabaseNotConnected", "Bad_Disabled", "Bad_Failure",
	  	"Bad_GatewayCommOff", "Bad_LicenseExceeded", "Bad_NotConnected", "Bad_NotFound", "Bad_OutOfRange", "Bad_ReadOnly",
	  	"Bad_ReferenceNotFound", "Bad_Stale", "Bad_TrialExpired", "Bad_Unauthorized", "Bad_Unsupported"]},
	  {"name": "Error", "exacts": ["Error_Configuration", "Error_CycleDetected", "Error_DatabaseQuery", "Error_Exception", "Error_ExpressionEval",
	  	"Error_Formatting", "Error_IO", "Error_InvalidPathSyntax", "Error_ScriptEval", "Error_TagExecution", "Error_TimeoutExpired", "Error_TypeConversion"]},
	  {"name": "Uncertain", "exacts": ["Uncertain_DataSubNormal", "Uncertain_EngineeringUnitsExceeded", "Uncertain_IncompleteOperation",
	  	"Uncertain_InitialValue", "Uncertain_LastKnownValue"]}
    ],
  "returnables":
    [
      {"alias": "Path", "type": "property", "key": "path"},
      {"alias": "Tag Type", "type": "property", "key": "tagType"},
      {"alias": "Quality", "type": "property", "key": "quality"},
      {"alias": "Alarm Eval Enabled", "type": "property", "key": "alarmEvalEnabled"},
      {"alias": "Alarms", "type": "property", "key": "alarms"},
      {"alias": "Clamp Mode", "type": "property", "key": "clampMode"},
      {"alias": "Data Type", "type": "property", "key": "dataType"},
      {"alias": "Datasource", "type": "property", "key": "datasource"},
      {"alias": "Deadband", "type": "property", "key": "deadband"},
      {"alias": "Deadband Mode", "type": "property", "key": "deadbandMode"},
      {"alias": "Deadband Style", "type": "property", "key": "historicalDeadbandStyle"},
      {"alias": "Documentation", "type": "property", "key": "documentation"},
      {"alias": "Enabled", "type": "property", "key": "enabled"},
      {"alias": "Engineering High Limit", "type": "property", "key": "engHigh"},
      {"alias": "Engineering Limit Mode", "type": "property", "key": "engLimitMode"},
      {"alias": "Engineering Low Limit", "type": "property", "key": "engLow"},
      {"alias": "Engineering Units", "type": "property", "key": "engUnit"},
      {"alias": "Execution Mode", "type": "property", "key": "executionMode"},
      {"alias": "Execution Rate (ms)", "type": "property", "key": "executionRate"},
      {"alias": "Expression", "type": "property", "key": "expression"},
      {"alias": "Format String", "type": "property", "key": "formatString"},
      {"alias": "Full Path", "type": "property", "key": "fullPath"},
      {"alias": "Historical Deadband", "type": "property", "key": "historicalDeadband"},
      {"alias": "Historical Deadband Mode", "type": "property", "key": "historicalDeadbandMode"},
      {"alias": "History Tag Group", "type": "property", "key": "historyTagGroup"},
      {"alias": "History Enabled", "type": "property", "key": "historyEnabled"},
      {"alias": "Max Time Between Samples", "type": "property", "key": "historyMaxAge"},
      {"alias": "Max Time Units", "type": "property", "key": "historyMaxAgeUnits"},
      {"alias": "Min Time Between Samples", "type": "property", "key": "historyTimeDeadband"},
      {"alias": "Min Time Units", "type": "property", "key": "historyTimeDeadbandUnits"},
      {"alias": "Name", "type": "property", "key": "name"},
      {"alias": "OPC Item Path", "type": "property", "key": "opcItemPath"},
      {"alias": "OPC Server", "type": "property", "key": "opcServer"},
      {"alias": "Overrides", "type": "property", "key": "tagQueryOverrides"},
      {"alias": "Parameters", "type": "property", "key": "parameters"},
      {"alias": "Parent Data Type", "type": "property", "key": "typeId"},
      {"alias": "Qualified Value", "type": "property", "key": "tagQueryQualifiedValue"},
      {"alias": "Query", "type": "property", "key": "query"},
      {"alias": "Query Type", "type": "property", "key": "queryType"},
      {"alias": "Raw High", "type": "property", "key": "rawHigh"},
      {"alias": "Raw Low", "type": "property", "key": "rawLow"},
      {"alias": "Read Expression", "type": "property", "key": "deriveExpressionGetter"},
      {"alias": "Read Only", "type": "property", "key": "readOnly"},
      {"alias": "Read Permissions", "type": "property", "key": "readPermissions"},
      {"alias": "Sample Mode", "type": "property", "key": "sampleMode"},
      {"alias": "Sample Rate", "type": "property", "key": "historySampleRate"},
      {"alias": "Sample Rate Units", "type": "property", "key": "historySampleRateUnits"},
      {"alias": "Scale Factor", "type": "property", "key": "scaleFactor"},
      {"alias": "Scale Mode", "type": "property", "key": "scaleMode"},
      {"alias": "Scaled High", "type": "property", "key": "scaledHigh"},
      {"alias": "Scaled Low", "type": "property", "key": "scaledLow"},
      {"alias": "Source Tag Path", "type": "property", "key": "sourceTagPath"},
      {"alias": "Storage Provider", "type": "property", "key": "historyProvider"},
      {"alias": "Tag Event Scripts", "type": "property", "key": "eventScripts"},
      {"alias": "Tag Group", "type": "property", "key": "tagGroup"},
      {"alias": "Tag ID", "type": "property", "key": "tagId"},
      {"alias": "Timestamp", "type": "property", "key": "tagquerytimestamp"},
      {"alias": "Tooltip", "type": "property", "key": "tooltip"},
      {"alias": "Type Color", "type": "property", "key": "typeColor"},
      {"alias": "Value", "type": "property", "key": "value"},
      {"alias": "Value Source", "type": "property", "key": "valueSource"},
      {"alias": "Write Expression", "type": "property", "key": "deriveExpressionSetter"},
      {"alias": "Write Permissions", "type": "property", "key": "writePermissions"},
      {"alias": "Alarm(s) Configured", "type": "trait", "key": "alarm"},
      {"alias": "Custom Security Permissions", "type": "trait", "key": "security"},
      {"alias": "Event Script Configured", "type": "trait", "key": "scripting"},
      {"alias": "History Enabled", "type": "trait", "key": "history"},
      {"alias": "Overrides Parent Properties", "type": "trait", "key": "override"},
      {"alias": "Tag Disabled", "type": "trait", "key": "disabled"},
      {"alias": "Value Scaling Applied", "type": "trait", "key": "scaling"}
    ],
    'propOps': [
      {'comp': 'Equal', 'visual': '=', 'useValue': True},
      {'comp': 'Not_Equal', 'visual': '!=', 'useValue': True},
      {'comp': 'Like', 'visual': 'Like', 'useValue': True},
      {'comp': 'Not_Like', 'visual': '!Like', 'useValue': True},
      {'comp': 'Defined', 'visual': 'Has', 'useValue': False},
      {'comp': 'Not_Defined', 'visual': '!Has', 'useValue': False},
      {'comp': 'Overridden', 'visual': 'Override', 'useValue': False},
      {'comp': 'GreaterThan', 'visual': '>', 'useValue': True},
      {'comp': 'GreaterThanEqual', 'visual': '>=', 'useValue': True},
      {'comp': 'LessThan', 'visual': '<', 'useValue': True},
      {'comp': 'LessThanEqual', 'visual': '<=', 'useValue': True}
    ]
}

query = {
	u'condition': {
		u'path': u'*',
		u'attributes': {
			u'values': [],
			u'requireAll': True
		}
	},
	u'returnProperties': ['path', 'tagType', 'quality'],
	u'options': {
		u'includeUdtMembers': True,
		u'includeUdtDefinitions': False
	}
}

historyColumns = [
	{
      "alias": "Path",
      "key": "path",
      "type": "property",
      "visible": True,
      "width": 200
    },
    {
      "alias": "Storage Provider",
      "key": "historyProvider",
      "type": "property",
      "visible": True,
      "width": 200
    }
]

historyQuery = {
	'condition': {
		'path': '*',
		'attributes': {
			'values': ['history'],
			'requireAll': False
		}
	},
	'returnProperties': ['path', 'historyProvider', 'alarms', 'engUnit', 'formatString'],
	'options': {
		'includeUdtMembers': True,
		'includeUdtDefinitions': False
	}
}

mixedHistoryAndAlarmColumns = [
	{
      "alias": "Path",
      "key": "path",
      "type": "property",
      "visible": True,
      "width": 200
    },
    {
      "alias": "Storage Provider",
      "key": "historyProvider",
      "type": "property",
      "visible": True,
      "width": 200
    },
    {
      "alias": "Alarms",
      "key": "alarms",
      "type": "property",
      "visible": True,
      "width": 200
    }
]

mixedHistoryAndAlarmQuery = {
	'condition': {
		'path': '*',
		'attributes': {
			'values': ['alarm', 'history'],
			'requireAll': False
		},
		"properties": {
		    "conditions": [
		        {
		            "conditions": [{
		                "comp": "Defined",
		                "prop": "alarms",
		                "value": ""
		            }],
	                "op": "And"
		        },
		        {
		            "conditions": [{
		                "comp": "Defined",
		                "prop": "historyProvider",
		                "value": ""
		            }],
		            "op": "And"
		        }
		    ],
		    "op": "Or"
		}
	},
	'returnProperties': ['path', 'historyProvider', 'alarms', 'engUnit', 'formatString'],
	'options': {
		'includeUdtMembers': True,
		'includeUdtDefinitions': False
	}
}

# Strip qualified values out of hierarchical entities.
def deep(x):
	if x is None:
		return None
	if isinstance(x, QualifiedValue):
		return deep(x.value)
	if isinstance(x, basestring):
		return x
	if isinstance(x, mappings):
		x = dict(x)
		retv = dict()
		for k, v in x.items():
			retv[k] = deep(v)
		return retv
	try:
		return [deep(v) for v in x]
	except:
		return x
#

# Run the tag query with a limit.
def runQuery(provider, query, limit):
	try:
		q = deep(query)
		if provider and query:
			logger.debugf("Running tag preview with provider=%s q=%s", provider, repr(q))
			if limit>0:
				wrapper = system.tag.query(provider, q, limit)
			else:
				wrapper = system.tag.query(provider, q)
			retv = dict(
				results=list(wrapper.results),
				continuation=wrapper.continuationPoint,
				availableSize=wrapper.totalAvailableSize,
				returnedSize=wrapper.returnedSize,
				provider=provider,
				query=q
			)
			alnumCI = Comparators.alphaNumeric(False)
			for d in retv['results']:
				d['path'] = d['fullPath'].toStringPartial()
				d['provider'] = d['fullPath'].getSource()
			retv['results'].sort(lambda x, y: alnumCI.compare(x['path'], y['path']))
		else:
			logger.debugf("Skipping tag preview with provider=%s q=%s", provider, repr(q))
			retv = None
		vvm = system.perspective.viewVarMap()
		vvm.tagQueryResult = retv
		vvm.refresh()
		return retv
	except Throwable, t:
		logger.warn("Unexpected tag query java error", t)
		raise t
	except Exception, e:
		logger.warn("Unexpected tag query jython error", exchange.tagReportUtility.later.PythonAsJavaException(e))
		raise e
#

# Handle click on the items in the qualities filter tree.  Special handling
# to make the hierarchy icons follow the selections:
# 1) When clicking on an outer item, the inner items are cleared and then
#    that part of the tree collapses.
# 2) When clicking on an inner item, the outer item is set to indeterminate
#    (unconditionally) to keep those items expanded.
def qualityClick(self, event):
	qmap = dict(self.view.custom.qualitiesMap)
	state = qmap.get(event.data, 0)
	parts = event.data.split('_')
	if len(parts)>1:
		prefix = parts[0]
		qmap[prefix] = 2
		qmap[event.data] = 0 if state else 1
	else:
		qmap[event.data] = 0 if state else 1
		possibles = []
		for x in self.custom.possibleQualities:
			if x.name == event.data:
				for k in x.exacts:
					qmap.pop(k, None)
				break
	self.props.selection = []
	self.view.custom.qualitiesMap = qmap
#

# Handle click on the items in the outputs and preview list.  While
# a tree component, there are no nested elements.  The icons are
# driven by the view.custom.outputsMap object, and clicking rotates
# through 0, 1, and 2.  Zero is no output nor preview.  One is output
# and preview.  Two is output without preview.
#
# Each item's data contains the map key.
def outputsClick(self, event):
	omap = dict(self.view.custom.outputsMap)
	if event.data == 'path':
		omap[event.data] = 1
	else:
		omap[event.data] = (omap.get(event.data, 0) + 1) % 3
	self.props.selection = []
	self.view.custom.outputsMap = omap
#

# Handle click on the items in the attributes filter list.  While
# a tree component, there are no nested elements.  The icons are
# driven by the view.custom.traitsMap object, and clicking alternates
# between 0 and 1.
#
# Each item's data contains the map key.
def traitsClick(self, event):
	tmap = dict(self.view.custom.traitsMap)
	tmap[event.data] = (tmap.get(event.data, 0) + 1) % 2
	self.props.selection = []
	self.view.custom.traitsMap = tmap
#

emptyPropSelect = {'prop': '', 'comp': 'Equal', 'value': ''}

# Add an outer (OR) property selector with an empty inner entry.
def addPropertyAny(self, event):
	self.view.custom.properties_OR_AND.append([emptyPropSelect])
#

# Handle a message from an embedded property filtering view to
# manipulate the current filter conditions.
def propsFilterMsg(self, payload):
	copy = deep(self.view.custom.properties_OR_AND)
	logger.debugf("Received props-filter payload %s, current conditions: %s, copy=%s",
		repr(payload),
		repr(self.view.custom.properties_OR_AND),
		repr(copy))
	if payload.get('closer', '') == self.view.params.closer:
		action = payload.get('action', '')
		index0 = payload.get('index0', -1)
		index1 = payload.get('index1', -1)
		if action == 'add':
			addPropertyAll(self, index0)
		elif action == 'delete':
			delProperty(self, index0, index1)
		elif action == 'update':
			prop = payload.get('key')
			value = payload.get('value')
			if prop:
				updateProperty(self, index0, index1, prop, value)
			else:
				logger.infof("Unexpected missing prop (%s) for value (%s)", prop, value)
		else:
			logger.infof("Unknown action: %s", action)
#

# Construct a base payload and action for the above message handler.
def baseEmbedPayload(self, action):
	payload = dict([(k,v) for (k, v) in dict(self.view.params).items() if k in ('closer', 'index0', 'index1')])
	payload['action'] = action
	return payload
#

# Add an empty inner (AND) property selector to the current outer entry.
# The event source is expected to have a view parameter "index" that
# identifies the outer entry.
def addPropertyAll(self, index0):
	self.view.custom.properties_OR_AND[index0].append(emptyPropSelect)
#

# Post a property filter add message from the outer filter's add button.
def sendPropertyAdd(self):
	payload = baseEmbedPayload(self, 'add')
	system.perspective.sendMessage('props-filter', payload, scope='page')
#

#
def delProperty(self, index0, index1):
	qty = len(self.view.custom.properties_OR_AND[index0])
	try:
		del self.view.custom.properties_OR_AND[index0][index1]
	except:
		return
	if qty > 1:
		return
	try:
		del self.view.custom.properties_OR_AND[index0]
	except:
		pass
#

# Post a property filter delete from the inner filter's delete button.
def sendPropertyDelete(self):
	payload = baseEmbedPayload(self, 'delete')
	system.perspective.sendMessage('props-filter', payload, scope='page')
#

# Assign a new value into a condition.
def updateProperty(self, index0, index1, prop, value):
	logger.debugf("Setting %s=%s @ %d/%d", prop, value, index0, index1)
	self.view.custom.properties_OR_AND[index0][index1][prop] = value
#

# Post a property filter update from within an inner filter component.
def sendPropertyUpdate(self, prop, value):
	payload = baseEmbedPayload(self, 'update')
	payload['key'] = prop
	payload['value'] = value
	system.perspective.sendMessage('props-filter', payload, scope='page')
#

# Handle a settings download request.  Extract the settings information from the
# last preview run.
def settingsDownloadRequest(self, event):
	content = dict(query=deep(self.view.custom.query), columns=deep(self.view.custom.columns))
	system.perspective.download("TagSelectorQuery.json", system.util.jsonEncode(content, 2))
#

# Handle a settings file received.  Interpret the payload as JSON and ensure
# it has a "query" key and optional "columns" key.  Apply them to the view
# parameters as if opened with the given settings.
def settingsUploadReceived(self, event):
	pass
#

# Opens the nested settings clipboard/database transfer popup.  That popup will
# let the user select from the user's available saved settings files.
def openSettings(self, event):
	uuid = str(UUID.randomUUID())
	system.perspective.openPopup(uuid, 'Exchange/TagReportUtility/SelectorPopups/Settings', {'query': deep(self.view.custom.query), 'columns': deep(self.view.custom.columns)})
	self.view.custom.popupTargetUUID = uuid
#

#
def openSettingsSave(self, event):
	pass
#

# Conditionally extend to the full result set.
def deliverableResults(self):
	vvm = system.perspective.viewVarMap(self)
	tqr = vvm.tagQueryResult
	if self.view.custom.selections:
		selected = set(self.view.custom.selections)
		pruned = dict(
			results=[tag for tag in tqr.get('results', []) if tag['path'] in selected],
			continuation=tqr.get('continuation', ''),
			availableSize=tqr.get('availableSize', 0),
			returnedSize=tqr.get('returnedSize', 0),
			provider=tqr.get('provider', ''),
			query=tqr.get('query', {})
		)
		tqr = pruned
	elif tqr['continuation']:
		runQuery(tqr['provider'], tqr['query'], 0)
		tqr = vvm.tagQueryResult
	tqr['columns'] = deep(self.view.custom.columns)
	return tqr
#

# Force complete query.
def previewAllResults(self, event):
	deliverableResults(self)
#

# Handle a results download
def downloadResults(self, event):
	outputName = "%s_TagSelect_%s.xlsx" % (system.util.getProjectName(), system.date.format(system.date.now(), 'yyyyMMdd_HHmmss'))
	retv = deliverableResults(self)
	wb, hStyle, rStyle, dStyle, jStyle = exchange.tagReportUtility.excel.createWorkbook()

	# Export content
	sheet0 = wb.createSheet('Selected Tags')
	headerRow = sheet0.createRow(0)
	columns = deep(retv['columns'])
	for c, colInfo in enumerate(columns):
		cell = headerRow.createCell(c)
		cell.setCellStyle(hStyle)
		cell.setCellValue(colInfo['alias'])
	doAutoSize = True
	for r, rowData in enumerate(retv['results']):
		row = sheet0.createRow(r+1)
		for c, colInfo in enumerate(columns):
			cell = row.createCell(c)
			value = rowData.get(colInfo['key'])
			if isinstance(value, Date):
				cell.setCellStyle(dStyle)
			else:
				cell.setCellStyle(rStyle)
			cell.setCellValue(value)
		if r == 50:
			# When more than fifty rows, do the autosize early, as it
			# can be very time-consuming.
			doAutoSize = False
			for c, colInfo in enumerate(columns):
				sheet0.autoSizeColumn(c)
	if doAutoSize:
		for c, colInfo in enumerate(columns):
			sheet0.autoSizeColumn(c)

	# Export configuration JSON
	sheet1 = wb.createSheet('Configuration JSON')
	row = sheet1.createRow(0)
	cell = row.createCell(0)
	cell.setCellStyle(jStyle)
	jsonStr = system.util.jsonEncode(dict(query=retv.get('query'), columns=retv.get('columns')), 2)
	cell.setCellValue(jsonStr)
	sheet1.autoSizeColumn(0)
	row.setHeightInPoints(1.15 * jStyle.font.fontHeightInPoints * len(jsonStr.splitlines()))

	exchange.tagReportUtility.excel.downloadWorkbook(wb, outputName)

# Handle a results delivery
def deliverResults(self, event):
	payload = dict(deliverableResults(self))
	supplement = deep(self.view.params.msg)
	handler = supplement.pop('handler', 'exchange.tagReportUtility.TagSelect')
	scope = supplement.pop('scope', 'page')
	payload.update(supplement)
	system.perspective.sendMessage(handler, payload, scope)
	#system.perspective.closePopup(self.view.params.closer)
#

# Open View popups/TagSelect as a modal to select tags based
# on a starting query.
def openTagSelect(self, provider, query, columns, handler='exchange.tagReportUtility.TagSelect', title='Select Tags...', replyKey=None, **kwargs):
	if replyKey is None:
		replyKey = str(UUID.randomUUID())
		self.view.custom.replyKey = replyKey
	closer = self.page.id + '/tagselect'
	params = dict(kwargs)
	params.update({
		'closer': closer,
		'provider': provider,
		'query': query,
		'columns': columns,
		'msg': {'handler': handler, 'scope': 'page', 'key': replyKey}
	})
	logger.debugf("Tag Select Popup parameters=%s", repr(params))
	system.perspective.openPopup(closer, 'Exchange/TagReportUtility/TagSelector', params, title, modal=True, resizable=True, viewportBound=True)
#
#
# kate: tab-width 4; indent-width 4; tab-indents on; dynamic-word-wrap off; indent-mode python; line-numbers on;
