# Tag Report Utility time span UI scripts.

# Copyright 2024 Automation Professionals, LLC <sales@automation-pros.com>
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#   1. Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#   2. Redistributions in binary form must reproduce the above copyright notice,
#      this list of conditions and the following disclaimer in the documentation
#      and/or other materials provided with the distribution.
#   3. The name of the author may not be used to endorse or promote products
#      derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
# SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
# OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
# OF SUCH DAMAGE.

from java.util import Date
from java.lang import System
from java.time import Instant, ZoneId, DayOfWeek
from java.time.format import DateTimeFormatter
from java.time.temporal import ChronoUnit

logger = system.util.getLogger(system.util.getProjectName()+'.'+system.reflect.getModulePath())

defaultFormatter = DateTimeFormatter.ofPattern("YYYY-MM-dd HH:mm:ss.SSSZ")

def zonedFormat(ts, zoneid, format=None):
	formatter = defaultFormatter if format is None else DateTimeFormatter.ofPattern(format)
	return ts.toInstant().atZone(ZoneId.of(zoneid)).format(formatter)

offsetFormatter = DateTimeFormatter.ofPattern('Z')

def zoneOffsetNow(zoneid):
	return Date().toInstant().atZone(ZoneId.of(zoneid)).format(offsetFormatter)

# Conditionally coerce Perspective dates that are long integer milliseconds.
def coerceDate(ts):
	if ts is None:
		return Date(0L)
	if isinstance(ts, Date):
		return ts
	return Date(ts)
#

# Truncate timestamps to a specific interval in milliseconds.
# Cannot safely be used for intervals greater than one hour, and
# cannot be safely used for intervals greater than one minute if the
# time zone offset has non-zero minutes.
def truncateSimpleTS(ts, modulus):
	millis = ts.time
	return Date(millis - millis % modulus)
#

def truncateToMidnight(ts, zoneid):
	zts = ts.toInstant().atZone(ZoneId.of(zoneid))
	midnight = zts.withHour(0).withMinute(0).withSecond(0).withNano(0).toInstant()
	return Date(midnight.toEpochMilli())
#

# Truncate timestamps to a specific granularity, accounting for shift start times.
def truncateToGranularity(ts, zoneid, granularity, startMonth, startDoW, startHour, startMinute):
	try:
		div, inc, modulus = granularity['divisor'], granularity['increment'], granularity['modulus']
	except:
		logger.warnf("Failed to unpack granularity=%s", repr(granularity))
		raise
	zts = ts.toInstant().atZone(ZoneId.of(zoneid))
	toLog = "%s w/ div=%s, inc=%s, mod=%s, stMonth=%s, stHour=%s, stMinute=%s =>" % (
		zts.format(defaultFormatter), div, inc, modulus, startMonth, startHour, startMinute
	)
	if modulus and modulus <= 3600000L:
		retv = truncateSimpleTS(ts, modulus)
		logger.infof("%s %s", toLog, retv)
		return retv
	# All below use day start.  Get start of the given day.
	midnightZDT = zts.withHour(0).withMinute(0).withSecond(0).withNano(0)
	dayStartZDT = midnightZDT.withHour(startHour).withMinute(startMinute)
	
	if modulus:
		# Eight or Twelve-hour shifts.  Watch out for time change
		dsMillis = dayStartZDT.toInstant().toEpochMilli()
		offset = dsMillis - midnightZDT.toInstant().toEpochMilli()
		adjustment = (modulus + ts.time - dsMillis) % modulus
		toLog += " ofs=%s, adj=%s" % (offset, adjustment)
		retv = zts.minus(adjustment, ChronoUnit.MILLIS)
	elif inc == 'Weeks':
		fromWeekStart = (7 + dayStartZDT.dayOfWeek.value - startDoW) % 7
		if fromWeekStart:
			retv = dayStartZDT.minusDays(fromWeekStart)
		else:
			retv = dayStartZDT
	elif inc == 'Months':
		retv = dayStartZDT.withDayOfMonth(1)
	elif inc == 'Quarters':
		retv = dayStartZDT.withDayOfMonth(1).withMonth(startMonth+1)
		if retv.isAfter(zts):
			while retv.isAfter(zts):
				retv = retv.minusMonths(3)
		else:
			candidate = retv.addMonths(3)
			while not candidate.isAfter(zts):
				retv = candidate
				candidate = retv.addMonths(3)
	elif inc == 'Years':
		retv = dayStartZDT.withDayOfMonth(1).withMonth(startMonth+1)
		if retv.isAfter(zts):
			retv = retv.minusYears(1)
	else:
		retv = dayStartZDT

	logger.infof("%s %s", toLog, retv)
	return Date(retv.toInstant().toEpochMilli())
#
#

# Update precise start, possibly bump precise end.
def alterStart(self, newTS):
	zoneid = self.session.props.timeZoneId
	cust = self.view.custom
	granularity = cust.bounds
	truncated = truncateToGranularity(newTS, zoneid, granularity, cust.qtrStartMonth, cust.weekStartDay, cust.dayStartHour, cust.dayStartMinute)
	self.view.params.preciseStart = truncated
	if not truncated.before(self.view.params.preciseEnd):
		div, inc, modulus = granularity['divisor'], granularity['increment'], granularity['modulus']
		addIncrement = getattr(system.date, 'add'+inc)
		self.view.params.preciseEnd = addIncrement(truncated, div)
	self.view.custom.stale = True
#

# Update precise end, possibly bump precise start.
def alterEnd(self, newTS):
	zoneid = self.session.props.timeZoneId
	cust = self.view.custom
	granularity = cust.bounds
	truncated = truncateToGranularity(newTS, zoneid, granularity, cust.qtrStartMonth, cust.weekStartDay, cust.dayStartHour, cust.dayStartMinute)
	self.view.params.preciseEnd = truncated
	if not truncated.after(self.view.params.preciseStart):
		div, inc, modulus = granularity['divisor'], granularity['increment'], granularity['modulus']
		addIncrement = getattr(system.date, 'add'+inc)
		self.view.params.preciseStart = addIncrement(truncated, -div)
	self.view.custom.stale = True
#

# Adjust timestamp to given date
def adjustDate(self, oldTS, newTS):
	zone = ZoneId.of(self.session.props.timeZoneId)
	newZDT = coerceDate(newTS).toInstant().atZone(zone)
	oldZDT = coerceDate(oldTS).toInstant().atZone(zone)
	adjusted = oldZDT.withYear(newZDT.year).withMonth(newZDT.monthValue).withDayOfMonth(newZDT.dayOfMonth)
	return Date(adjusted.toInstant().toEpochMilli())
#

# Adjust timestamp to given hour, minute, second
def adjustTime(self, oldTS, timeParts):
	zone = ZoneId.of(self.session.props.timeZoneId)
	oldZDT = coerceDate(oldTS).toInstant().atZone(zone)
	parts = exchange.tagReportUtility.selector.deep(timeParts)
	adjusted = oldZDT.withHour(parts.get('hour', 0)).withMinute(parts.get('minute', 0)).withSecond(parts.get('second', 0)).withNano(0)
	return Date(adjusted.toInstant().toEpochMilli())
#

# Adjust timestamp to given hour
def adjustHour(self, oldTS, newHour):
	zone = ZoneId.of(self.session.props.timeZoneId)
	oldZDT = coerceDate(oldTS).toInstant().atZone(zone)
	adjusted = oldZDT.withHour(newHour).withNano(0)
	return Date(adjusted.toInstant().toEpochMilli())
#

# Adjust timestamp to given minute
def adjustMinute(self, oldTS, newMinute):
	zone = ZoneId.of(self.session.props.timeZoneId)
	oldZDT = coerceDate(oldTS).toInstant().atZone(zone)
	adjusted = oldZDT.withMinute(newMinute).withNano(0)
	return Date(adjusted.toInstant().toEpochMilli())
#

# Adjust timestamp to given second
def adjustSecond(self, oldTS, newSecond):
	zone = ZoneId.of(self.session.props.timeZoneId)
	oldZDT = coerceDate(oldTS).toInstant().atZone(zone)
	adjusted = oldZDT.withSecond(newSecond).withNano(0)
	return Date(adjusted.toInstant().toEpochMilli())
#

# Force precise start and end to fall on the required boundaries.
# "self" must be the TimeSpan view.
def alterBounds(self):
	zoneid = self.session.props.timeZoneId
	cust = self.view.custom
	granularity = cust.bounds
	oldStart = coerceDate(self.view.params.preciseStart)
	newStart = truncateToGranularity(oldStart, zoneid, granularity, cust.qtrStartMonth, cust.weekStartDay, cust.dayStartHour, cust.dayStartMinute)
	oldEnd = coerceDate(self.view.params.preciseEnd)
	newEnd = truncateToGranularity(oldEnd, zoneid, granularity, cust.qtrStartMonth, cust.weekStartDay, cust.dayStartHour, cust.dayStartMinute)
	div, inc, modulus = granularity['divisor'], granularity['increment'], granularity['modulus']
	addIncrement = getattr(system.date, 'add'+inc)
	minEnd = addIncrement(newStart, div)
	if minEnd.after(newEnd):
		newEnd = minEnd
	self.view.params.preciseStart = newStart
	self.view.params.preciseEnd = newEnd
	self.view.custom.stale = True
#
#
# kate: tab-width 4; indent-width 4; tab-indents on; dynamic-word-wrap off; indent-mode python; line-numbers on;
